#!/usr/bin/env python

'''A check plugin to monitor the status of "stuff" in op5Live'''

author = 'Joel Rangsmo <jrangsmo@op5.com>'
url = 'https://kb.op5.com/x/egHU'
version = '0.5'

# Don't edit these for testing - use "-E' and "-P" arguments
endpoint_address = 'api.op5live.com'
endpoint_port = 443

try:
    from op5_monitor_api_client import MonitorSession
    import argparse
    import logging
    import json
    import time

except ImportError as missing:
    print (
        'Error - could not import all required Python modules: "%s"\n\n'
        % missing + 'Dependency installation with pip:\n'
        '"# pip install op5_monitor_api_client argparse"\n\n'
        'Dependency installation on with yum:\n'
        '"# yum install python-op5_monitor_api_client python-argparse"\n')

    exit(3)


# Argument parsing
aparser = argparse.ArgumentParser(
    description=__doc__,
    epilog='Developed by %s - For more information see: "%s"'
    % (author, url))

aparser.add_argument(
    '-u', '--user',
    help='Specify op5Live user', required=True)

aparser.add_argument(
    '-T', '--token',
    help='Specify op5Live API access token', required=True)

# You could either use the plugin to monitor host/service or list available
check_or_list = aparser.add_mutually_exclusive_group(required=True)

check_or_list.add_argument(
    '-c', '--check-id',
    help='Check ID in op5Live to monitor')

check_or_list.add_argument(
    '-q', '--query-mode',
    help=argparse.SUPPRESS, action='store_true', default=False)

# Used to specify API endpoint - hidden by default, but useful for testing
aparser.add_argument(
    '-E', '--endpoint-address',
    help=argparse.SUPPRESS, default=endpoint_address)

aparser.add_argument(
    '-P', '--endpoint-port',
    help=argparse.SUPPRESS, type=int, default=endpoint_port)

# General plugin options
aparser.add_argument(
    '-d', '--debug',
    help='Execute plugin in debug mode',
    action='store_true', default=False)

aparser.add_argument(
    '-v', '--version',
    help='Display plugin version',
    action='version', version=version)

args = aparser.parse_args()


# Exits the check plugin with appropriate output
def exit_plugin(state, output, data):

    # Dumps JSON data if plugin was used in query mode
    if args.query_mode:
        print json.dumps({
                'query_status': state,
                'query_message': output,
                'query_data:': data})

        exit(state)

    else:
        print output
        exit(state)


# Enables debug logging to stream if debug mode is enabled
if args.debug:
    logger = logging.basicConfig(level=logging.DEBUG)


# Builds the query list filter
if args.check_id:
    query_filter = (
            '[services] host.name = "%s" or description = "%s"'
            % (args.check_id, args.check_id))

else:
    query_filter = '[services] all'


# Columns requested in the API query (in the "services" category)
query_columns = [
        'host.name', 'host.display_name', 'host.state', 'host.plugin_output',
        'description', 'display_name', 'state', 'plugin_output',
        'host.perf_data_raw', 'perf_data_raw']

# Creates the op5Live session object
op5live = MonitorSession(
        username=args.user, password=args.token, version=7.04,
        server=args.endpoint_address, port=args.endpoint_port)

# The amout of time it takes for the query against op5Live - used in perfdata
query_start = time.time()

try:
    query = op5live.query_filter(
            lfilter=query_filter, columns=query_columns)

except MonitorSession.ConnectionError:
    exit_plugin(
            3, 'Failed to retrive data from op5Live endpoint (%s:%i)'
            % (args.endpoint_address, args.endpoint_port), '')

except MonitorSession.AuthError:
    exit_plugin(
            3, 'Username and/or API access token was invalid', '')

except MonitorSession.APIError as error:
    exit_plugin(
            3, 'An unknown issue occured while querying op5Live - '
            'Debug information: "%s"' % error, '')

query_time = str(time.time() - query_start)

# If plugin is used in query mode, a list of available checks will be JSONed
if args.query_mode:
    exit_plugin(
            0, 'Host and service query was successful', query)


# Checks that the name or description is acctually unique and exists
if not len(query):
    exit_plugin(
            3, 'Could not find check ID in op5Live (see "%s")' % url, '')

elif len(query) > 1:
    exit_plugin(
            3, 'Issue occured while querying op5Live (ID was not unique)', '')

query = query[0]

# Checks if the check ID was for host or service status
if query['host']['name'] == args.check_id:
    check = query['host']

elif query['description'] == args.check_id:
    check = query

else:
    exit_plugin(
            3, 'Unknown issue occured while parsing op5Live status query', '')


# Includes the remote checks perfdata and the op5Live API query time
perfdata = (
    " | 'op5Live query time'=%ss %s" % (query_time, check['perf_data_raw']))

# Exits the plugin with the op5Live checks status, output and query time
exit_plugin(
        check['state'], 'op5Live: ' + check['plugin_output'] + perfdata, '')
