#!/usr/bin/env python

'''op5_monitor_api_client.py - a Python module and
command line tool for communication with op5 Monitor\'s HTTP API'''

version = '0.1'
description = __doc__
author = 'Joel Rangsmo <jrangsmo@op5.com>'
license = 'LGPLv2'

# Enables CLI mode if application is used stand alone
if __name__ == '__main__':
    cli_mode = True

else:
    cli_mode = False

# Imports required Python modules
try:
    import logging
    import logging.handlers
    import requests
    import json

    # Using the application in CLI mode requires additional dependencies
    if cli_mode:
        import argparse

except ImportError as missing:
    if cli_mode:
        print (
            'Error - could not import all required Python modules: "%s"\n'
            % missing + 'Dependency installation with pip: '
            '"# pip install argparse requests json"')

        exit(3)

    else:
        raise ImportError(missing)


logger = logging.getLogger('op5_monitor_api_client')


# Object used for communicating with the op5 Monitor REST API
class MonitorSession(object):
    '''Object used for communicating with the op5 Monitor REST API'''

    # Exception objects
    class APIError(Exception):
        '''All exceptions inherits from this one'''

        def __init__(self, msg):
            self.msg = msg

        def __str__(self):
            return self.msg

    # Connection and session related exceptions
    class ConnectionError(APIError):
        '''Exceptions related connection issues'''

    class VerificationError(ConnectionError):
        '''Exception for invalid/self-signed HTTPS certificates'''

    class ResponseError(ConnectionError):
        '''Exception for invalid/self-signed HTTPS certificates'''

    class AuthError(APIError):
        '''Exceptions related to invalid or missing credentials'''

    class PrivError(APIError):
        '''Exceptions related to insufficient privileges'''

    # Query related errors
    class QueryError(APIError):
        '''Exceptions related to issues with queries'''

    # / Exception objects

    def __init__(self, server=None, username=None, password=None, **kwargs):
        self.server = server
        self.username = username
        self.password = password

        # Checks if mandatory arguments are provided
        if not self.server or not self.username or not self.password:
            raise ValueError(
                'Argument "server", "username" '
                'or "password" was not provided')

	self.timeout = kwargs.get('timeout', 20)
        self.port = kwargs.get('port', 443)
        self.verify = kwargs.get('verify', True)
        self.path = kwargs.get('path', '/api')
        self.version = kwargs.get('version', 7.05)

        logger.debug(
            'Created API communication object with '
            'username "%s" for op5 Monitor server "%s" port %i. '
            % (self.username, self.server, self.port) +
            'Certificate verification enabled: %s. ' % str(self.verify) +
            'API version compatibility: "%s"' % str(self.version) +
            'API path: "%s". ' % (self.path))

    # Function to get data from Monitor's API
    def get_data(self, subpath, options):
        '''Function to get data from Monitor\'s API.
        Expects two arguments - a API path and URL options in a dict'''

	# Sets format to JSON
	options['format'] = 'json'

        logger.debug(
            'Getting data from path "%s" with URL options "%s%s"'
            % (self.path, subpath, options))

        try:
            query = requests.get(
                'https://%s:%i%s/%s'
                % (self.server, self.port, self.path, subpath),
		verify=self.verify, timeout=self.timeout,
                auth=(self.username, self.password), params=options)

        # Exception handling (timeouts, SSL issues, etc)
        except requests.exceptions.Timeout:
            raise self.ConnectionError(
                'Connection to server "%s" timed out'
                % self.server)

        except requests.exceptions.SSLError:
            raise self.VerificationError(
                'Failed to verify certificate for server "%s"'
                % self.server)

        except requests.exceptions.ConnectionError:
            raise self.ConnectionError(
                'Failed to connect to server "%s"'
                % self.server)

        except requests.exceptions.RequestException as unknown:
            raise self.APIError(
                'Unknown connection exception: "%s"' % unknown)

        logger.debug(
            'Server response data: status code: "%s", text: "%s"'
            % (query.status_code, query.text))

        return query

    # Function to check status codes from Monitor's API responses
    def check_status_code(self, response):
        '''Function to check common status codes from Monitor\'s API.
        Expects one argument - a object generated by requests,
        including status code and response data'''

        logger.debug(
            'Server responded with status code %i - response text: "%s"'
            % (response.status_code, response.text))

        if response.status_code == 200:
            logger.debug('Status code in server response was OK')

            return True

        if response.status_code == 201:
            logger.debug('Status code in server response was OK')

            return True

        elif response.status_code == 400:
            raise self.QueryError(
                'Request was not understood by server - are you sure the '
                'server is running op5 Monitor %s or later?' % self.version)

        elif response.status_code == 401:
            raise self.AuthError(
                'Invalid username, password for user "%s" on server "%s"'
                % (self.username, self.server))

        elif response.status_code == 403:
            raise self.PrivError(
                'Insufficient privileges for user "%s" on server "%s"'
                % (self.username, self.server))

        elif response.status_code == 404:
            raise self.QueryError(
                'Server responded with file not found - are you '
                'sure the server is running op5 Monitor 6.3 or later?')

        else:
            raise self.ResponseError(
                'Server responded with unknown status code: %i. '
                % response.status_code + 'Response text: "%s"'
                % response.text)

        return False

    # Function to decode JSON data in response to Python list
    def decode_response(self, response):
        '''Function to decode JSON data in response to Python list.
        Expects one argument - a JSON blob, hopefully containing data'''

        logger.debug('Decoding JSON response data')

        try:
            response = json.loads(response)

        except ValueError:
            raise self.ResponseError(
                'Failed to decode data in response from filter query')

        logger.debug('Response was decoded successfully')

        return response

    # Function to query list filters
    def query_filter(self, lfilter=None, **kwargs):
        '''Function to query list filters.
        Expects two arguments - a filter query and a list of filter columns.
        Returns list with dicts (decoded JSON response)'''

        # Checks if mandatory arguments are provided
	if not lfilter:
            raise ValueError('Argument "lfilter" was not provided')

	columns = kwargs.get('columns', None)	
	limit = kwargs.get('limit', 999999999)	

        logger.debug(
            'Querying filter "%s" for columns "%s"' % (lfilter, columns))

        # Joining columns to expected format and building URL options
        if columns and not isinstance(columns, list):
            raise self.QueryError('Filter columns were not provided as list')

        options = {
            'query': lfilter, 'limit': limit}

	if columns:
            options['columns'] = ','.join(columns)

        # Performs the filter query and validates the response
        query = self.get_data('filter/query', options)

        if self.check_status_code(query):
            return self.decode_response(query.text)

        else:
            return False

    # Function to count items found by filter qyery
    def count_filter(self, lfilter):
        '''Function to count items found by filter query.
        Expects one argument - a filter query.
        Returns a integer with filter count'''

        logger.debug(
            'Counting items found by filter query "%s"' % lfilter)

        options = {'query': lfilter}

        # Performs the filter query and validates the response
        query = self.get_data('filter/count', options)

        if self.check_status_code(query):
            query = self.decode_response(query.text)

            try:
                return query['count']

            except KeyError:
                raise self.ResponseError(
                    'Recieved invalid or malformed server response')


# Parses command line arguments
def arg_parser(description, author, version):
    '''Parses command line arguments.
    The function takes three arguments - decription, version and author.
    These will be used in the generation of the help text'''

    # op5 Monitor API session options
    parser = argparse.ArgumentParser(
        description=description,
        epilog='Developed by %s - Version %s' % (author, version))

    parser.add_argument(
        '-s', '--server',
        help='Specifies op5 Monitor server host (default: 127.0.0.1)',
        default='127.0.0.1')

    parser.add_argument(
        '-u', '--username',
        help='Specifies username for op5 Monitor user',
        required=True)

    parser.add_argument(
        '-p', '--password',
        help='Specifies password for op5 Monitor user',
        required=True)

    parser.add_argument(
        '-m', '--monitor-version', dest='mon_version',
        help='Specifies version of op5 Monitor (default: 7.0.4)',
        type=float, default=7.04)

    parser.add_argument(
        '-i', '--insecure', dest='verify',
        help='Disable certificate verification (not recommended)',
        action='store_false', default=True)

    # General application options
    parser.add_argument(
        '-l', '--logging',
        help='Set internal application logging destination',
        choices=('stream', 'syslog', 'none'), default='stream')

    parser.add_argument(
        '-V', '--verbose',
        help='Enable verbose logging',
        action='store_true', default=False)

    parser.add_argument(
        '-v', '--version', help='Display application version',
        action='version', version=version)

    args = parser.parse_args()

    return args


# Configures application logging
def log_init(destination, verbose):
    '''Configures application logging.
    The function takes two arguments - logging destination
    (Syslog, stream or none) and if debug output should be enabled'''

    logger = logging.getLogger('op5_monitor_api_client')
    formatter = logging.Formatter(
        'op5_monitor_api_client: %(levelname)s - %(message)s')

    if verbose:
        logger.setLevel(logging.DEBUG)

    else:
        logger.setLevel(logging.INFO)

    if destination == 'stream':
        loghandler = logging.StreamHandler()

    elif destination == 'syslog':
        loghandler = logging.handlers.SysLogHandler(address='/dev/log')

    elif destination == 'none':
        loghandler = logging.NullHandler()

    loghandler.setFormatter(formatter)
    logger.addHandler(loghandler)

    logger.debug(
        'Logging has been initiated - destined for %s' % destination)

    return logger


# Runs application in CLI mode
def main(description, author, version):
    '''Runs application in CLI mode.
    The function takes three arguments - description, author and version.'''

    # Parses provided command line arguments
    args = arg_parser(description, author, version)

    # Configures application logging
    global logger
    logger = log_init(args.logging, args.verbose)

    print 'op5_monitor_api_client can not be used in CLI mode ATM :('

    exit(3)


# Runs main function if program is being used stand alone
if cli_mode:
    main(description, author, version)
